@@ -296,11 +296,6 @@ namespace xt
         return trivial_broadcast;
     }
 
-    /***************************************
-     * xexpression_assigner implementation *
-     ***************************************/
-
-
     /********************************
      * data_assigner implementation *
      ********************************/
@@ -364,7 +359,7 @@ namespace xt
         using size_type = typename E1::size_type;
         size_type size = e1.size();
         size_type simd_size = simd_type::size;
-        size_type align_begin = is_aligned ? 0 : xsimd::get_alignment_offset(&(e1.data()), size, simd_size);
+        size_type align_begin = is_aligned ? 0 : xsimd::get_alignment_offset(e1.raw_data(), size, simd_size);
         size_type align_end = align_begin + ((size - align_begin) & ~(simd_size - 1));
         for (size_type i = 0; i < align_begin; ++i)
         {
