@@ -100,11 +100,11 @@ namespace xt
         {                                                                        \
             return (arg1 OP arg2);                                               \
         }                                                                        \
-        template <class U>                                                      \
-        struct rebind                                                           \
-        {                                                                       \
-            using type = NAME<U>;                                               \
-        };                                                                      \
+        template <class U>                                                       \
+        struct rebind                                                            \
+        {                                                                        \
+            using type = NAME<U>;                                                \
+        };                                                                       \
     }
 
 #define BINARY_OPERATOR_FUNCTOR(NAME, OP) BINARY_OPERATOR_FUNCTOR_IMPL(NAME, OP, T)
@@ -119,10 +119,11 @@ namespace xt
         BINARY_OPERATOR_FUNCTOR(minus, -);
         BINARY_OPERATOR_FUNCTOR(multiplies, *);
         BINARY_OPERATOR_FUNCTOR(divides, /);
+        BINARY_OPERATOR_FUNCTOR(modulus, %);
         BINARY_BOOL_OPERATOR_FUNCTOR(logical_or, ||);
         BINARY_BOOL_OPERATOR_FUNCTOR(logical_and, &&);
         UNARY_BOOL_OPERATOR_FUNCTOR(logical_not, !);
-        BINARY_OPERATOR_FUNCTOR(bitwise_or, | );
+        BINARY_OPERATOR_FUNCTOR(bitwise_or, |);
         BINARY_OPERATOR_FUNCTOR(bitwise_and, &);
         BINARY_OPERATOR_FUNCTOR(bitwise_xor, ^);
         UNARY_OPERATOR_FUNCTOR(bitwise_not, ~);
@@ -363,6 +364,23 @@ namespace xt
         return detail::make_xfunction<detail::divides>(std::forward<E1>(e1), std::forward<E2>(e2));
     }
 
+    /**
+     * @ingroup arithmetic_operators
+     * @brief Modulus
+     *
+     * Returns an \ref xfunction for the element-wise modulus
+     * of \a e1 by \a e2.
+     * @param e1 an \ref xexpression or a scalar
+     * @param e2 an \ref xexpression or a scalar
+     * @return an \ref xfunction
+     */
+    template <class E1, class E2>
+    inline auto operator%(E1&& e1, E2&& e2) noexcept
+    -> detail::xfunction_type_t<detail::modulus, E1, E2>
+    {
+        return detail::make_xfunction<detail::modulus>(std::forward<E1>(e1), std::forward<E2>(e2));
+    }
+    
     /**
      * @defgroup logical_operators Logical operators
      */
@@ -675,8 +693,7 @@ namespace xt
         auto idx = xtl::make_sequence<index_type>(arr.dimension(), 0);
         std::vector<index_type> indices;
 
-        auto next_idx = [&shape](index_type& idx)
-        {
+        auto next_idx = [&shape](index_type& idx) {
             for (size_type j = shape.size(); j > 0; --j)
             {
                 size_type i = j - 1;
@@ -773,7 +790,7 @@ namespace xt
     /**
      * @defgroup casting_operators Casting operators
      */
-    
+
     /**
      * @ingroup casting_operators
      * @brief Element-wise ``static_cast``.
@@ -784,14 +801,13 @@ namespace xt
      * @param e an \ref xexpression or a scalar
      * @return an \ref xfunction
      */
-    
+
     template <class R, class E>
     inline auto cast(E&& e) noexcept
-    -> detail::xfunction_type_t<detail::cast<R>::template functor, E>
+        -> detail::xfunction_type_t<detail::cast<R>::template functor, E>
     {
         return detail::make_xfunction<detail::cast<R>::template functor>(std::forward<E>(e));
     }
-
 }
 
 #endif
